import uuid, os
from pathlib import Path
from xdagtool import xdu


class TestPathCopy:
    def test_file_copy1(self, tmp_path: Path, capsys):
        # file a
        srcpath = tmp_path / 'srcpath-{}'.format(uuid.uuid4())
        destpath = tmp_path / 'destpath-{}'.format(uuid.uuid4())

        with capsys.disabled():
           print(tmp_path)

        srcpath.mkdir(parents=True)
        destpath.mkdir(parents=True)

        tmpfname = 'a-{}'.format(uuid.uuid4())
        tmp_txt = 'text-{}'.format(uuid.uuid4())
        srcfile_path = srcpath / tmpfname
        srcfile_path.write_text(tmp_txt)
        destfile_path = destpath / tmpfname

        xdu.path_copy(srcfile_path, destfile_path)

        assert destfile_path.exists()
        assert destfile_path.read_text() == tmp_txt

    def test_file_copy2(self, tmp_path: Path, capsys):
        # file a/b/c
        srcpath = tmp_path / 'd1'
        destpath = tmp_path / 'd2'

        srcpath.mkdir(parents=True)
        destpath.mkdir(parents=True)

        tmpfile = Path('c') / 'b' / 'a-{}'.format(uuid.uuid4())
        tmp_txt = 'text-{}'.format(uuid.uuid4())
        srcfile_path = srcpath / tmpfile
        srcfile_path.parent.mkdir(parents=True, exist_ok=True)
        srcfile_path.write_text(tmp_txt)
        destfile_path = destpath / tmpfile

        xdu.path_copy(srcfile_path, destfile_path)

        assert destfile_path.exists()
        assert destfile_path.read_text() == tmp_txt

    # @pytest.mark.skip(reason='temp')
    def test_dir_copy1(self, tmp_path: Path, capsys):
        # dir srcpath/1.dat 2.dat
        srcpath = tmp_path / 'd1'
        destpath = tmp_path / 'd2'

        srcpath.mkdir(parents=True)

        # src dir file list
        tmp_txt1 = 'text-{}'.format(uuid.uuid4())
        srcfile1 = srcpath / 'a-{}'.format(uuid.uuid4())
        srcfile1.write_text(tmp_txt1)

        tmp_txt2 = 'text-{}'.format(uuid.uuid4())
        srcfile2 = srcpath / 'a-{}'.format(uuid.uuid4())
        srcfile2.write_text(tmp_txt2)

        # go
        xdu.path_copy(srcpath, destpath)

        # assert
        assert destpath.exists()

        destfile1 = destpath / srcfile1.name
        assert destfile1.exists()
        assert destfile1.read_text() == tmp_txt1

        destfile2 = destpath / srcfile2.name
        assert destfile2.exists()
        assert destfile2.read_text() == tmp_txt2

    def test_dir_copy2(self, tmp_path: Path, capsys):
        # dir srcpath/ empty
        srcpath = tmp_path / 'd1'
        destpath = tmp_path / 'd2'

        srcpath.mkdir(parents=True)

        # src dir is empty

        # go
        xdu.path_copy(srcpath, destpath)

        # assert
        assert destpath.exists()
        assert len(os.listdir(destpath)) == 0

    def test_dir_copy3(self, tmp_path: Path, capsys):
        # dir srcpath/b/c/ empty
        srcpath = tmp_path / 'd1' / 'b' / 'c'
        destpath = tmp_path / 'd2' / 'b' / 'c'

        srcpath.mkdir(parents=True)

        # src dir is empty

        # go
        xdu.path_copy(srcpath, destpath)

        # assert
        destdir = tmp_path / 'd2'
        assert destdir.exists()

        destdir = tmp_path / 'd2' / 'b'
        assert destdir.exists()

        destdir = tmp_path / 'd2' / 'b' / 'c'
        assert destdir.exists()

        assert len(os.listdir(destdir)) == 0

    def test_dir_copy4(self, tmp_path: Path, capsys):
        # dir srcpath/b/c/ 1.dat 2.dat
        srcpath = tmp_path / 'd1' / 'b' / 'c'
        destpath = tmp_path / 'd2' / 'b' / 'c'

        srcpath.mkdir(parents=True)

        # src dir files
        tmp_txt1 = 'text-{}'.format(uuid.uuid4())
        srcfile1 = srcpath / 'a-{}'.format(uuid.uuid4())
        srcfile1.write_text(tmp_txt1)

        tmp_txt2 = 'text-{}'.format(uuid.uuid4())
        srcfile2 = srcpath / 'a-{}'.format(uuid.uuid4())
        srcfile2.write_text(tmp_txt2)

        # go
        xdu.path_copy(srcpath, destpath)

        # assert
        destdir = tmp_path / 'd2'
        assert destdir.exists()

        destdir = tmp_path / 'd2' / 'b'
        assert destdir.exists()

        destdir = tmp_path / 'd2' / 'b' / 'c'
        assert destdir.exists()

        assert destpath.exists()

        destfile1 = destpath / srcfile1.name
        assert destfile1.exists()
        assert destfile1.read_text() == tmp_txt1

        destfile2 = destpath / srcfile2.name
        assert destfile2.exists()
        assert destfile2.read_text() == tmp_txt2

    def test_dir_copy5(self, tmp_path: Path, capsys):
        # dir srcpath/b/
        #               c1/d1/1.dat 2.dat       # dir1
        #               c2/1.dat 2.dat          # dir2
        #               c3/                     # dir3
        srcpath = tmp_path / 'd1' / 'b'
        destpath = tmp_path / 'd2' / 'b'

        srcpath.mkdir(parents=True)

        # src dir initialize
        # dir1
        dir1 = srcpath / 'c1' / 'd1'
        dir1.mkdir(parents=True)

        dir1_tmp_txt1 = 'text-{}'.format(uuid.uuid4())
        dir1_srcfile1 = dir1 / 'a-{}'.format(uuid.uuid4())
        dir1_srcfile1.write_text(dir1_tmp_txt1)

        dir1_tmp_txt2 = 'text-{}'.format(uuid.uuid4())
        dir1_srcfile2 = dir1 / 'a-{}'.format(uuid.uuid4())
        dir1_srcfile2.write_text(dir1_tmp_txt2)

        # dir2
        dir2 = srcpath / 'c2'
        dir2.mkdir(parents=True)

        dir2_tmp_txt1 = 'text-{}'.format(uuid.uuid4())
        dir2_srcfile1 = dir2 / 'a-{}'.format(uuid.uuid4())
        dir2_srcfile1.write_text(dir2_tmp_txt1)

        dir2_tmp_txt2 = 'text-{}'.format(uuid.uuid4())
        dir2_srcfile2 = dir2 / 'a-{}'.format(uuid.uuid4())
        dir2_srcfile2.write_text(dir2_tmp_txt2)

        # dir3
        dir3 = srcpath / 'c3'
        dir3.mkdir(parents=True)


        # go
        xdu.path_copy(srcpath, destpath)

        # assert
        assert destpath.exists()

        # dir1
        destdir = destpath / 'c1' / 'd1'
        assert destdir.exists()

        destfile1 = destdir / dir1_srcfile1.name
        assert destfile1.exists()
        assert destfile1.read_text() == dir1_tmp_txt1

        # dir2
        destdir = destpath / 'c2'
        assert destdir.exists()

        # dir3
        destdir = destpath / 'c3'
        assert destdir.exists()


class TestPathsCopy:
    def test_paths_copy1(self, tmp_path: Path, shared_datadir: Path, capsys):
        assert (shared_datadir / 'storage').exists()
        xdu.paths_copy(str(shared_datadir / 'storage'), str(tmp_path / 'testdata'), ['01/75/f4/18.dat',
                                                                                     '01/75/f4/32.dat',
                                                                                     '01/75/f4/32.dat',
                                                                                     '01/75/f4/40.dat',
                                                                                     '01/75/f4/42.dat',  # not exist
                                                                                     '01/75/f4/sums.dat',
                                                                                     '01/75/f7',
                                                                                     '01/75/f8',    # not exist
                                                                                     '01/75/f9',
                                                                                     'empty1',
                                                                                     'empty2/empty3',])
        assert (tmp_path / 'testdata/01/75/f4/40.dat').exists()
        assert (tmp_path / 'testdata/01/75/f4/32.dat').exists()
        assert (tmp_path / 'testdata/01/75/f4/sums.dat').exists()
        assert (tmp_path / 'testdata/01/75/f7/01.dat').exists()
        assert (tmp_path / 'testdata/01/75/f9/sums.dat').exists()
        assert (tmp_path / 'testdata/empty1').exists()
        assert (tmp_path / 'testdata/empty2/empty3').exists() and (tmp_path / 'testdata/empty2/empty3').is_dir()

        assert not (tmp_path / 'testdata/01/75/f4/39.dat').exists()
        assert not (tmp_path / 'testdata/01/75/fb').exists()

        assert (not (tmp_path / 'storage/01/75/f4/42.dat').exists()) and (not (tmp_path / 'testdata/01/75/f4/42.dat').exists())
        assert (not (tmp_path / 'storage/01/75/f8').exists()) and (not (tmp_path / 'testdata/01/75/f8').exists())



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="xdagtool", # Replace with your own username
    version="0.4",
    author="Larry Wu",
    author_email="rulerson@qq.com",
    description="Xdag slicing tool",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/rulerson/xdagtool",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.7',
    install_requires=['python-dateutil', 'click'],
    keywords='xdag',
    entry_points = {
        'console_scripts': ['xdag-slice=xdagtool.xdu:cmd_go'],
    },
)



import click
from dateutil import parser
from pathlib import Path
import shutil


g_verbose: bool = False

def time_slice(time_start, time_end):
    time_start_ts = int(time_start.timestamp())
    time_end_ts = int(time_end.timestamp())

    time_start_slice = time_start_ts >> 6
    time_end_slice = time_end_ts >> 6

    if g_verbose:
        print('time_start_ts={0}, time_start_hex={0:X}, time_start_slice={1:X}, time_end_ts={2}, '
              'time_end_hex={2:X}, time_end_slice={3:X}'.format(
            time_start_ts, time_start_slice, time_end_ts, time_end_slice))

    start_level_4 = time_start_slice & 0xFF
    start_level_3 = (time_start_slice >> 8) & 0xFF
    start_level_2 = (time_start_slice >> 16) & 0xFF
    start_level_1 = (time_start_slice >> 24) & 0xFF

    end_level_4 = time_end_slice & 0xFF
    end_level_3 = (time_end_slice >> 8) & 0xFF
    end_level_2 = (time_end_slice >> 16) & 0xFF
    end_level_1 = (time_end_slice >> 24) & 0xFF

    if g_verbose:
        print('start_level={:x}-{:x}-{:x}-{:x}, end_level={:x}-{:x}-{:x}-{:x}'.format(
            start_level_1, start_level_2, start_level_3, start_level_4,
            end_level_1, end_level_2, end_level_3, end_level_4))
        print('start_level={}-{}-{}-{}, end_level={}-{}-{}-{}'.format(
            start_level_1, start_level_2, start_level_3, start_level_4,
            end_level_1, end_level_2, end_level_3, end_level_4))

    return time_start_slice, time_end_slice


class SlicePath:
    def __init__(self, time_start_slice, time_end_slice, debug=False):
        self.__time_start_slice = time_start_slice
        self.__time_end_slice = time_end_slice
        self.__debug = debug
        self.__paths = []

    def paths(self):
        self.__iter_level1()
        return self.__paths

    def __iter_level1(self):
        for x in range(0, 0xFF + 1):
            xv_min = x << 24
            xv_max = xv_min + 0xFFFFFF
            level_path = '{}'.format(x) if self.__debug else '{:0>2x}'.format(x)

            if xv_max < self.__time_start_slice or xv_min > self.__time_end_slice:
                continue

            if xv_min > self.__time_start_slice and xv_max < self.__time_end_slice:
                self.__paths.append('{}/'.format(level_path))
            else:
                self.__iter_level2(xv_min, level_path)

    def __iter_level2(self, level_start, parent_path):
        for x in range(0, 0xFF + 1):
            xv_min = x << 16
            xv_max = xv_min + 0xFFFF
            level_path = '{}/{}'.format(parent_path, x) if self.__debug else '{}/{:0>2x}'.format(parent_path, x)

            if level_start + xv_max < self.__time_start_slice or level_start + xv_min > self.__time_end_slice:
                continue

            if level_start + xv_min > self.__time_start_slice and level_start + xv_max < self.__time_end_slice:
                self.__paths.append('{}/'.format(level_path))
            else:
                self.__iter_level3(level_start + xv_min, level_path)

    def __iter_level3(self, level_start, parent_path):
        for x in range(0, 0xFF + 1):
            xv_min = x << 8
            xv_max = xv_min + 0xFF
            level_path = '{}/{}'.format(parent_path, x) if self.__debug else '{}/{:0>2x}'.format(parent_path, x)

            if level_start + xv_max < self.__time_start_slice or level_start + xv_min > self.__time_end_slice:
                continue

            if level_start + xv_min > self.__time_start_slice and level_start + xv_max < self.__time_end_slice:
                self.__paths.append('{}/'.format(level_path))
            else:
                self.__iter_level4(level_start + xv_min, level_path)

    def __iter_level4(self, level_start, parent_path):
        for x in range(0, 0xFF + 1):
            level_path = '{}/{}.dat'.format(parent_path, x) if self.__debug else '{}/{:0>2x}.dat'.format(parent_path, x)

            if level_start + x >= self.__time_start_slice and level_start + x <= self.__time_end_slice:
                self.__paths.append('{}'.format(level_path))


def path_copy(src_path: Path, dest_path: Path):
    # mkdir -p
    dest_path.parent.mkdir(parents=True, exist_ok=True)

    # copy go
    src_path_name, dest_path_name = str(src_path), str(dest_path)
    if src_path.is_dir():
        if g_verbose:
            print('copy dir from {} to {}'.format(src_path_name, dest_path_name))
        shutil.copytree(src_path_name, dest_path_name)
    else:
        if g_verbose:
            print('copy file from {} to {}'.format(src_path_name, dest_path_name))
        shutil.copy2(src_path_name, dest_path_name)


def paths_copy(srcdir: str, destdir: str, paths):
    srcdir_path = Path(srcdir).resolve()
    destdir_path = Path(destdir).resolve() if Path(destdir).is_absolute() else (Path.cwd() / Path(destdir)).resolve()
    print('copy go, srcdir_p={}, destdir_p={}'.format(srcdir_path, destdir_path))

    for path_v in paths:
        file_path = Path(path_v)
        src_path = srcdir_path / file_path
        dest_path = destdir_path / file_path

        # skip no-exist file/dir
        if not src_path.exists():
            continue

        # copy go
        path_copy(src_path, dest_path)



@click.command(context_settings=dict(help_option_names=['-h', '--help']))
@click.option('-s', '--src-dir', 'srcdir', default='.', show_default=True,
              help='source directory copy from, default to current directory.',
              type=click.Path(exists=True, file_okay=False, writable=True, resolve_path=True), )
@click.option('-d', '--dst-dir', 'dstdir', default=None, show_default=True,
              help='dest directory copy to, default to dryrun mode if not supplied.',
              type=click.Path(exists=True, file_okay=False, writable=True, resolve_path=True))
@click.option('-r', '--dryrun', is_flag=True,
              help='dryrun mode, show all files and directories prepare to process, but do not really run.',
              default=False, show_default=True)
@click.option('-v', '--verbose', is_flag=True,
              help='verbose mode',
              default=False, show_default=True)
@click.argument('time-start', type=str)
@click.argument('time-end', type=str, required=False, default=None)
def cmd_go(srcdir, dstdir, dryrun, time_start, time_end, verbose):
    """xdag blocks slicing.

    <command> start-time end-time

        show all files and directories prepare to process between start-time and end-time.

        star-time, end-time, iso8601 format.

    <command> time

        find target block filename for specific time slice. (start-time=end-time=time.)

        time iso8601 format.

    <command> -s srcdir -d dstdir start-time end-time

        copy all files and directories between start-time and end-time from srcdir to dstdir.

    notes:

        xdag main-net era time: 2018-01-05T22:45:00

        lack of sums.dat for some directorys.

        iso8601 time format examples (https://dateutil.readthedocs.io/en/stable/parser.html#dateutil.parser.isoparse):

            2018-01-05T22:45:00Z, 2018-01-05T22:45:00+08:00, always use utc time or timezone suffix

    """

    # command parameters
    # print(srcdir, dstdir, dryrun, time_start, time_end)
    global g_verbose
    g_verbose = verbose

    if not dstdir:
        dryrun = True
    time_start_dt = parser.isoparse(time_start)
    time_end_dt = parser.isoparse(time_end) if time_end else time_start_dt
    print('time_start={}, time_end={}'.format(time_start_dt, time_end_dt))

    # get slices
    time_start_slice, time_end_slice = time_slice(time_start_dt, time_end_dt)

    # get paths
    paths = SlicePath(time_start_slice, time_end_slice, debug=False).paths()

    # dryrun or really run
    if dryrun:      # only print paths list
        print('Dryrun mode')
        for x in paths:
            print(x)
    else:           # copy
        paths_copy(srcdir, dstdir, paths)

    return 0


if __name__ == '__main__':
    cmd_go()
